# coding=utf-8

"""Command line processing"""


import argparse
from snappysonic import __version__
from snappysonic.ui.snappysonic_demo import run_demo


def main(args=None):
    """Entry point for snappysonic application"""

    parser = argparse.ArgumentParser(
        description='snappysonic')

    ## ADD POSITIONAL ARGUMENTS
    parser.add_argument(
        "-c",
        "--config",
        required=True,
        type=str,
        help="A file containing the configuration."
        )


    version_string = __version__
    friendly_version_string = version_string if version_string else 'unknown'
    parser.add_argument(
        "--version",
        action='version',
        version='snappysonic version ' +
        friendly_version_string
        )

    args = parser.parse_args(args)

    run_demo(args.config)



#!/usr/bin/python
#  -*- coding: utf-8 -*-
"""snappysonic"""

import sys

if __name__ == "__main__" and not __package__:
    # To allow the package's main function to be executed without the -m switch,
    # i.e. "python snappysonic", we have to explicitly set the
    # module name and append the parent directory to the sys.path (see PEP 366)
    from os import path
    __package__ = "snappysonic"  # pylint: disable=redefined-builtin
    sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))
    sys.path.append(path.dirname(path.dirname(__file__)))

# pylint: disable=wrong-import-position
from snappysonic.ui.snappysonic_command_line import main

if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



"""Functions for snappysonic"""
from numpy import iinfo, int16
from PySide2.QtGui import QPixmap, QImage
from sksurgerynditracker.nditracker import NDITracker
from sksurgeryarucotracker.arucotracker import ArUcoTracker

def configure_tracker(config):
    """
    Configures a scikit-surgery tracker based on the passed config

    :param config: a tracker configuration dictionary
    :returns: The tracker
    :raises: KeyError
    """

    if "tracker type" not in config:
        raise KeyError('Tracker configuration requires tracker type')

    tracker_type = config.get("tracker type")
    tracker = None
    if tracker_type in ("vega", "polaris", "aurora", "dummy"):
        tracker = NDITracker(config)
    if tracker_type in "aruco":
        tracker = ArUcoTracker(config)

    if tracker_type not in "dummy":
        tracker.start_tracking()
    return tracker


def lookupimage(usbuffer, pts):
    """
    determines whether a coordinate (pts) lies with an area defined by
    a usbuffer, and returns an image from the buffer if appropriate

    :param usbuffer: a dictionary containing bounding box information (x0,y0,
        x1,y1) and image data

    :returns: True if point in bounding box. Image.
    """
    if pts[0] > usbuffer.get("x0"):
        if pts[0] < usbuffer.get("x1"):
            if pts[1] > usbuffer.get("y0"):
                if pts[1] < usbuffer.get("y1"):
                    pdiff = 0
                    if usbuffer.get("scan direction") == "x":
                        diff = pts[0] - usbuffer.get("x0")
                        pdiff = int(diff /
                                    (usbuffer.get("x1") - usbuffer.get("x0")) *
                                    len(usbuffer.get("buffer")))
                    else:
                        diff = pts[1] - usbuffer.get("y0")
                        pdiff = int(diff /
                                    (usbuffer.get("y1") - usbuffer.get("y0")) *
                                    len(usbuffer.get("buffer")))

                    return True, usbuffer.get("buffer")[pdiff]

    return False, None

def check_us_buffer(usbuffer):
    """
    Checks that all ultrasound buffer contains all required key values.

    :param usbuffer: the buffer to check
    :raises: KeyError
    :raises: ValueError
    """
    if "name" not in usbuffer:
        raise KeyError("Buffer configuration must contain a name.")
    if "start frame" not in usbuffer:
        raise KeyError("Buffer configuration must contain a start frame.")
    if "end frame" not in usbuffer:
        raise KeyError("Buffer configuration must contain an end frame.")
    if "x0" not in usbuffer:
        raise KeyError("Buffer configuration must contain x0")
    if "x1" not in usbuffer:
        raise KeyError("Buffer configuration must contain x1")
    if "y0" not in usbuffer:
        raise KeyError("Buffer configuration must contain y0")
    if "y1" not in usbuffer:
        raise KeyError("Buffer configuration must contain y1")
    if "scan direction" not in usbuffer:
        raise KeyError("Buffer configuration must contain a scan direction.")

    direction = usbuffer.get("scan direction")
    if direction not in ("x", "y"):
        raise ValueError("scan direction must be either x or y")

def get_bg_image_size(config):
    """
    Reads the geometry from a configuration and
    returns the extents of the buffer

    :param config: a tracker configuration dictionary
    :returns: the extents of the buffer
    """
    min_x = iinfo(int16).max
    max_x = iinfo(int16).min
    min_y = iinfo(int16).max
    max_y = iinfo(int16).min

    if "buffer descriptions" in config:
        for usbuffer in config.get("buffer descriptions"):
            if usbuffer.get("x0") > max_x:
                max_x = usbuffer.get("x0")
            if usbuffer.get("x1") > max_x:
                max_x = usbuffer.get("x1")
            if usbuffer.get("x1") < min_x:
                min_x = usbuffer.get("x1")
            if usbuffer.get("x0") < min_x:
                min_x = usbuffer.get("x0")
            if usbuffer.get("y0") > max_y:
                max_y = usbuffer.get("y0")
            if usbuffer.get("y1") > max_y:
                max_y = usbuffer.get("y1")
            if usbuffer.get("y1") < min_y:
                min_y = usbuffer.get("y1")
            if usbuffer.get("y0") < min_y:
                min_y = usbuffer.get("y0")
    border_size = 60
    if "border size" in config:
        border_size = config.get("border size")

    max_x = max_x + border_size
    min_x = min_x - border_size

    max_y = max_y + border_size
    min_y = min_y - border_size

    offsets = (min_x, min_y)
    image_size = (max_y - min_y, max_x - min_x)

    return offsets, image_size

def numpy_to_qpixmap(np_image):
    """
    Converts the input numpy array to a qpixmap

    :param np_image: a numpy array
    :returns: a qpixmap
    """
    height, width = np_image.shape
    q_image = QImage(np_image, width, height, QImage.Format_Grayscale8)
    return QPixmap.fromImage(q_image)



# coding=utf-8

"""Main loop for tracking visualisation"""
from packaging import version
from PySide2.QtWidgets import QLabel, QWidget
from cv2 import (rectangle, putText, circle, imread)
from numpy import zeros, uint8
from sksurgeryutils.common_overlay_apps import OverlayBaseWidget
from sksurgeryimage.utilities.weisslogo import WeissLogo
from sksurgeryarucotracker import __version__ as arucoversion
from snappysonic.algorithms.algorithms import (configure_tracker,
                                               lookupimage,
                                               check_us_buffer,
                                               get_bg_image_size,
                                               numpy_to_qpixmap)


class OverlayApp(OverlayBaseWidget):
    """
    Inherits from OverlayBaseWidget,
    adding code to read in video buffers, and display a frame
    of data that depends on the position of an external tracking system,
    e.g. surgeryarucotracker
    """

    def __init__(self, config):
        """
        Overides overlay base app's init, to initialise the
        external tracking system. Together with a video source

        :param config: a configuration dictionary
        """

        if "ultrasound buffer" in config:
            #and call the constructor for the base class
            super().__init__(config.get("ultrasound buffer"),
                    init_vtk_widget=False)
        else:
            raise KeyError("Configuration must contain an ultrasound buffer")

        self._video_buffers = self._fill_video_buffers(config)

        self._tracker = None

        if "tracker config" in config:
            self._tracker = configure_tracker(config.get("tracker config"))

        self._bgimage_offsets = (0, 0)
        self._backgroundimage = self._create_background_image(config)

        self._weiss = None
        self._defaultimage = zeros(0, dtype=uint8)
        if "default image" in config:
            self._defaultimage = imread(config.get("default image"))
        else:
            self._weiss = WeissLogo()

        self._logger = None

        self._tracking_window = QWidget()
        self._tracking_window.setWindowTitle("Tracker Position")

        self._tracking_viewer = QLabel("Tracking", self._tracking_window)

        height, width = self._backgroundimage.shape
        self._tracking_window.resize(width, height)
        self._tracking_viewer.resize(width, height)
        self._tracking_window.show()
        self._tracking_viewer.setPixmap(numpy_to_qpixmap(self._backgroundimage))

        default_target = 0
        if version.parse(arucoversion) >= version.parse("0.2.0"):
            default_target = "DICT_4X4_50:0"

        self._tracking_target = config.get("tracking target", default_target)
        #we could implement something like this?
        #if "log directory" in config:
        #    self._logger = sksurgerydatasaver(config.get("log directory"))

    def update_view(self):
        """
        Update the background renderer with a new frame,
        move the model and render
        """
        #add a method to move the rendered models
        image = self._get_image_with_tracking()

        self.vtk_overlay_window.set_video_image(image)
        self.vtk_overlay_window.Render()

    def _get_image_with_tracking(self):
        """
        Internal method to get an image from the video
        buffer based on the tracker position
        """
        port_handles, _, _, tracking, _ = self._tracker.get_frame()

        tempimg = self._backgroundimage.copy()

        pts = None
        if port_handles:
            for index, port_hdl in enumerate(port_handles):
                if port_hdl == self._tracking_target:
                    pts = (int(tracking[index][0, 3]),
                           int(tracking[index][1, 3]))
                    off_pts = (int(pts[0] - self._bgimage_offsets[0]),
                               int(pts[1] - self._bgimage_offsets[1]))
                    circle(tempimg, off_pts, 5, [255, 255, 255])

        self._tracking_viewer.setPixmap(numpy_to_qpixmap(tempimg))

        if pts:
            for usbuffer in self._video_buffers:
                inframe, image = lookupimage(usbuffer, pts)
                if inframe:
                    return image

        if self._defaultimage.shape == (0,):
            return self._weiss.get_noisy_logo()

        return self._defaultimage

    def _fill_video_buffers(self, config):
        """
        internal method to fill video buffers
        """
        vidbuffers = []
        frame_counter = 0
        if "buffer descriptions" in config:
            for usbuffer in config.get("buffer descriptions"):
                check_us_buffer(usbuffer)
                start_frame = usbuffer.get("start frame")
                end_frame = usbuffer.get("end frame")
                tempbuffer = []
                if start_frame == frame_counter:
                    while frame_counter <= end_frame:
                        ret, image = self.video_source.read()
                        if not ret:
                            raise ValueError("Failed Reading video file",
                                             config.get("ultrasound buffer"))

                        frame_counter = frame_counter + 1
                        tempbuffer.append(image)
                print("adding frame ", len(tempbuffer), " images to buffer ",
                      usbuffer.get("name"))
                usbuffer.update({"buffer" : tempbuffer})

                vidbuffers.append(usbuffer)

        return vidbuffers

    def _create_background_image(self, config):
        """
        Creates a backgound image on which we can draw tracking information.

        :param config: a configuration dictionary
        """

        bg_image_size = [480, 640]
        if "buffer descriptions" in config:
            self._bgimage_offsets, bg_image_size = get_bg_image_size(config)

        bgimage = zeros((bg_image_size), uint8)
        if "buffer descriptions" in config:
            for usbuffer in config.get("buffer descriptions"):
                pt0 = (usbuffer.get("x0") - self._bgimage_offsets[0],
                       usbuffer.get("y0") - self._bgimage_offsets[1])
                pt1 = (usbuffer.get("x1") - self._bgimage_offsets[0],
                       usbuffer.get("y1") - self._bgimage_offsets[1])
                rectangle(bgimage, pt0, pt1, [255, 255, 255])
                putText(bgimage, usbuffer.get("name"), pt0, 0,
                        1.0, [255, 255, 255])

        return bgimage



# coding=utf-8
"""
Setup for snappysonic 
"""

from setuptools import setup, find_packages
import versioneer

# Get the long description
with open('README.rst') as f:
    long_description = f.read()

setup(
    name='snappysonic',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='snappysonic provides an application to simulate an ultrasound probe moving over a body torso, it is intended for public engagement events"',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/SciKit-Surgery/snappysonic',
    author='Stephen Thompson',
    author_email='s.thompson@ucl.ac.uk',
    license='BSD-3 license',
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',


        'License :: OSI Approved :: BSD License',


        'Programming Language :: Python',
        'Programming Language :: Python :: 3',

        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
    ],

    keywords='medical imaging education',

    packages=find_packages(
        exclude=[
            'doc',
            'tests',
            'data'
        ]
    ),

    install_requires=[
        'opencv-contrib-python',
        'scikit-surgeryimage>=0.6.0',
        'scikit-surgeryutils>=1.2.1',
        'scikit-surgerynditracker',
        'scikit-surgeryarucotracker',
        'numpy',
        'PySide2',
    ],

    entry_points={
        'console_scripts': [
            'snappysonic=snappysonic.__main__:main',
        ],
    },
)



# coding=utf-8


"""SnappySonicdemo module"""

from PySide2.QtWidgets import QApplication
from sksurgerycore.configuration.configuration_manager import (
        ConfigurationManager
        )
from snappysonic.overlay_widget.overlay import OverlayApp


def run_demo(configfile):
    """ Run the application """

    configurer = ConfigurationManager(configfile)

    app = QApplication([])

    configuration = configurer.get_copy()

    viewer = OverlayApp(configuration)
    viewer.vtk_overlay_window.Initialize()
    viewer.vtk_overlay_window.Start()
    viewer.show()

    viewer.start()

    #start the application
    app.exec_()
    viewer.stop()
